"""Acoustic File-Transfer-Protocol Protocol-IDs"""
FTP_FILE_PID = 0xa0
#FTP_FILE_PID_LEN = 8

FTP_ACK_PID = 0xa1
#FTP_ACK_PID_LEN = 4
#FTP_ACK_BLOCK_LEN = 4

FTP_NET_PID = 0xa2
#FTP_NET_PID_LEN = 12

"""Begin prototype address space"""
FIRST_PROTO_PID = 0xa0
"""Header terminator PID"""
END_PID = 0xff
